<?php
namespace vimmaniac\mpl;

require_once(VMNC_MSH_PATH.'/lib/util/logging.php');
require_once(VMNC_MSH_PATH.'/lib/util/base_exception.php');

use vimmaniac\logging as logging;
if (! isset($VMNC_MPL_logger))
    { $VMNC_MPL_logger = new logging\CreateLogger('MPL'); }

$VMNC_MPL_task_prefetch_factor = 1.5;

$VMNC_MPL_fix_windows_quirk_1 = True;
$VMNC_MPL_fix_windows_quirk_2 = True;

if (! defined('VMNC_OS_IS_WINDOWS')){
    define(
        'VMNC_OS_IS_WINDOWS'
        , (strtolower(substr(PHP_OS, 0, 3)) == 'win')
    );
}

$VMNC_MPL_common_windows_envvars = array(
    'ALLUSERSPROFILE','APPDATA','CD','CMDCMDLINE','CMDEXTVERSION',
    'COMPUTERNAME','COMSPEC','DATE','ERRORLEVEL','HOMEDRIVE','HOMEPATH',
    'HOMESHARE','LOGONSERVER','NUMBER_OF_PROCESSORS','OS','PATH','PATHEXT',
    'PROCESSOR_ARCHITECTURE','PROCESSOR_IDENTIFIER','PROCESSOR_LEVEL',
    'PROCESSOR_REVISION','PROMPT','RANDOM','SYTEMDRIVE','SYSTEMROOT','TEMP',
    'TMP','TIME','USERDOMAIN','USERNAME','USERPROFILE','WINDIR'
);

interface
TaskSupplierInterface {
    # Method called before child processes are spawned.
    public function do_before_start_of_split();

    # Method called called after every child process has spawned.
    public function do_after_end_of_split();

    # Method called before child processes start to join.
    public function do_before_start_of_join();
    
    # Method called after every child process has joined.
    public function do_after_end_of_join();

    # Method should return an array containing information
    # necessary to perform the tasks it provides. As envvars
    # are used to pass this data, supply_task_hint should
    # return reasonably sized small amounts of data.
    public function supply_task_hint();

    # To overcome limitations of supply_task_hint, this method
    # should be used whenever large amounts of data needs to
    # be transfered between task supplier and task performer.
    public function supply_large_task_hint();

    # Method should return an array containing new tasks.
    # The number of tasks in the array should be close to
    # the $max_n_tasks parameter and must not exceed it.
    public function supply_new_tasks($max_n_tasks);

    # This method is called periodically. It must poll for finish/end
    # condition and if such conditions are met,
    # should raise  a 'UnnaturalShutdown' exception.
    public function periodic_callback_A();

    public function process_task_status($task_status);

    # This method receives the $ppm_finish_status produced by
    # the 'parent process manager' and should return a 'finish
    # status' convenient to the original problem space.
    public function modify_finish_status($ppm_finish_status);
}

interface
TaskPerformerInterface {
    # Any class implementing iTaskPerformer must accept
    # a 'task_hint' parameter and should store it internally
    # for later reference. This __construct() declaration serves
    # documentation purpose only.
    public function __construct($task_hint, $large_task_hint);

    # This method should be called from the constructor. It
    # prepares the object for performing the tasks given,
    # (such as setting up IPC) based on the 'task_hint' property.
    public function setup();

    # This method should execute tasks listed in the 'tasks' array,
    # sequentially. Ocassionally, it should call the function passed
    # as 'periodic_polling_callback' which polls for exit conditions
    # and raises proper exceptions when necessary.
    public function perform_tasks($tasks, $periodic_callback_B);

    # This method should be called from the destructor. It should
    # perform all necessary cleanups required for a clean exit.
    public function teardown();
}

# Base class for all exceptions used in this namespace.
class MplException extends \vimmaniac\BaseException {
    public function
    __construct(
        $msg=Null,
        $code=Null,
        \Exception $previous=Null,
        $log_level=VMNC_LOG_LEVEL_DEBUG
    ){
        global $VMNC_MPL_logger;

        parent::__construct($msg, $code, $previous);
        $VMNC_MPL_logger->add($this->__toString(), VMNC_LOG_LEVEL_DEBUG);
    }
}

# Something went wrong with 'Inter Process Communication'.
class IpcError extends MplException {
    public function
    __construct(
        $msg=Null,
        $code=Null,
        \Exception $previous=Null, 
        $log_level=VMNC_LOG_LEVEL_ERROR
    ){
        if ($msg === '') { $msg = "There was an IPC related error."; }
        parent::__construct($msg, $code, $previous, $log_level);
    }
}

######## Finalized Exceptions (Should not be inherited from) #########

final class EmptyAfterDuplicateCheck extends MplException {
    public function
    __construct(\Exception $previous=Null){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_WARN);
    }
}

final class RequestedToResign extends MplException {
    public function
    __construct(
        $final_task_status=Null,
        \Exception $previous=Null
    ){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_INFO);
        $this->task_status = $final_task_status;
    }
}

# 'task chunk pool' is empty. Time to get new tasks from the supplier.
final class EmptyTaskChunkPool extends MplException {}

# No more new tasks to supply.
final class TasksExhausted extends MplException {}

# Some other shutdown conditions were met, other than task exhaustion.
final class UnnaturalShutdown extends MplException {}

# Timeout occured while waiting for new task requests.
final class WaitForNewRequestTimedOut extends MplException {}

final class ThisShouldNeverHappenError extends MplException {}

# Failure occured while subprocessing the children.
final class FailedToStartSubprocess extends MplException {
    public function
    __construct(\Exception $previous=Null){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}

final class NotAChildProcess extends MplException {
    public function
    __construct(\Exception $previous=Null){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}

final class MissingTaskHint extends MplException {
    public function
    __construct(\Exception $previous=Null){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}

final class MissingProcessHint extends MplException {
    public function
    __construct(\Exception $previous=Null){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}

final class InvalidTaskHint extends MplException {
    public function
    __construct(\Exception $previous=Null){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}

final class InvalidProcessHint extends MplException {
    public function
    __construct(\Exception $previous=Null){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}
########################################################################

abstract class
ParentProcessManagerAbstract {
    public function
    __construct(
        # An instance of a class that implements 'iTaskSupplier' instance.
        $task_supplier_object,

        # The path of the php executable.
        $php_executable_path,

        # The path of the script which **USES** 'ChildProcessManager' class.
        $child_script_path,

        # Number of child processes.
        $n_children,

        # Size of array containing tasks, passed as responses
        # to 'task request's made by the children.
        $task_chunk_size,

        # The time delay between successive calls to 'do_periodic_polling'
        # method of the 'task_supplier_object'.
        $periodic_polling_interval,

        $prevent_task_duplication = True
    ){
        global $VMNC_MPL_logger;

        $this->task_supplier_object = $task_supplier_object;
        $this->php_executable_path = $php_executable_path;
        $this->child_script_path = $child_script_path;

        $this->n_children = $n_children;
        $this->task_chunk_size = $task_chunk_size;
        $this->periodic_polling_interval = $periodic_polling_interval;

        $this->prevent_task_duplication = $prevent_task_duplication;

        $VMNC_MPL_logger->add(
            sprintf(
                "Instance of class '%s' created with paremeters -"
                . " task_supplier_object: '''%s'''"
                . ", php_executable_path: '''%s'''"
                . ", child_script_path: '''%s'''"
                . ", n_children: '''%s'''"
                . ", task_chunk_size: '''%s'''"
                . ", periodic_polling_interval: '''%s'''"
                . ", prevent_task_duplication: '''%s'''"
                , get_class($this)
                , get_class($this->task_supplier_object)
                , $this->php_executable_path
                , $this->child_script_path
                , $this->n_children
                , $this->task_chunk_size
                , $this->periodic_polling_interval
                , $this->prevent_task_duplication
            )
            , VMNC_LOG_LEVEL_DEBUG
        );

        $this->child_list = array();
        $this->task_chunk_pool = array();
        $this->tasks_in_progress = array();

    }

    # The method should set and configure any 'Inter Process
    # Communication'(IPC) necessary to communicate with the children.
    abstract protected function
    setup_ipc();

    # The method sends the task chunk referred by 'task_chunk'
    # through IPC to the child identified by 'child_id'.
    abstract protected function
    send_task_chunk($child_id, $task_chunk);

    abstract protected function
    send_large_task_hint($child_id, $large_task_hint);

    abstract protected function
    send_resignment_letter_ack($child_id);

    abstract protected function
    ask_to_try_later($child_id);

    # This method signals the child through IPC to resign.
    abstract protected function
    tell_child_to_resign($child_id);

    # This method should return an array with information
    # necessary for the 'Child Process Manager' to function properly.
    abstract protected function
    supply_process_hint($child_id);

    # This method should monitor the IPC mechanism for incoming
    # requests by the children and return an array containing
    # 'child id's of those who requested for new tasks. This
    # method may raise a 'WaitForNewRequestTimedOut' exception if there is
    # no new task requests available within a timeout period.
    abstract protected function
    wait_for_new_request();

    # Should perform any cleanup action necessary on IPC mechanism,
    # before shutdown.
    abstract protected function
    teardown_ipc();

    protected function
    create_children(){
        for ($i = 0; $i < $this->n_children; $i++){
            # Create a child with 'id' equal to $i.
            $this->create_a_child((string)$i);
        }
    }

    protected function
    create_a_child($child_id){
        global
            $VMNC_MPL_logger
            , $VMNC_LOG_extra
            , $VMNC_MPL_common_windows_envvars
            , $VMNC_MPL_enable_windows_envvar_hack
            , $VMNC_MPL_fix_windows_quirk_1
            , $VMNC_MPL_fix_windows_quirk_2
        ;

        $VMNC_MPL_logger->add(
            "Spawning child-{$child_id} ..."
            , VMNC_LOG_LEVEL_INFO
        );

        $task_hint = $this->task_supplier_object->supply_task_hint();
        $proc_hint = $this->supply_process_hint($child_id);

        $env = array(
            'MPL_PROCESS_IS_CHILD' => '1'
            , 'MPL_TASK_HINT' => serialize($task_hint)
            , 'MPL_PROCESS_HINT' => serialize($proc_hint)
        );

        if (VMNC_OS_IS_WINDOWS && $VMNC_MPL_fix_windows_quirk_1){
            foreach ($VMNC_MPL_common_windows_envvars as $envname){
                if (($value = getenv($envname)) !== False){
                    $env[$envname] = $value;
                }
            }
        }

        $desc_spec = array();
        $cmd = $this->php_executable_path." -f ".$this->child_script_path;
        if (VMNC_OS_IS_WINDOWS && $VMNC_MPL_fix_windows_quirk_2)
            { $cmd = $cmd . " 2>&1 "; }

        $VMNC_MPL_logger->add(
            sprintf(
                "task_hint: '''%s'''"
                . ", process_hint: '''%s'''"
                . ", proc_open cmd: '''%s'''"
                , var_export($task_hint, True)
                , var_export($proc_hint, True)
                , $cmd
            )
            , VMNC_LOG_LEVEL_DEBUG
        );

        $process = proc_open($cmd, $desc_spec, $pipes, NULL, $env);
        if (! is_resource($process))
            { throw new FailedToStartSubprocess(); }

        $this->child_list[$child_id]['process'] = $process;
        $this->child_list[$child_id]['pipes'] = $pipes;
    }


    protected function
    resign_children($force=False){
        # Wait for children to finish.
        foreach(array_keys($this->child_list) as $child_id){
            $this->resign_a_child($child_id, $force);
        }
    }
    

    protected function
    resign_a_child($child_id, $force){
        global $VMNC_MPL_logger;

        $VMNC_MPL_logger->add(
            (($force)
                ? "Forcefully terminating child-{$child_id} ..."
                : "Waiting for child-{$child_id} to terminate ..."
            )
            , VMNC_LOG_LEVEL_INFO
        );

        $process = $this->child_list[$child_id]['process'];
        $pipes = $this->child_list[$child_id]['pipes'];
        
        # Close the pipes.
        foreach($pipes as $pipe) { fclose($pipe); }
        
        return (($force)
                ? proc_terminate($process)
                : proc_close($process));
    }


    protected function
    setup_clock(){
        $this->last_clock = time();
    }
    

    protected function
    check_clock(){
        $clock = time();

        if ($clock - $this->last_clock > $this->periodic_polling_interval){
            $this->last_clock = $clock;
            $this->task_supplier_object->periodic_callback_A();
        }
    }


    protected function
    assign_new_task_chunk($child_id){
        # Remove completed tasks from 'tasks_in_progress'.
        if (isset($this->child_list[$child_id]['task_chunk'])){
            $this->tasks_in_progress = array_diff(
                $this->tasks_in_progress,
                $this->child_list[$child_id]['task_chunk']
            );
        }

        # Grab a new 'task chunk' from the pool.
        $task_chunk = array_shift($this->task_chunk_pool);
        if ($task_chunk === NULL) { throw new EmptyTaskChunkPool(); }

        # Map the chunk against the particular child.
        $this->child_list[$child_id]['task_chunk'] = $task_chunk;

        # Update 'tasks_in_progress'.
        $this->tasks_in_progress =
            array_merge($this->tasks_in_progress, $task_chunk);

        # Return the newly assigned chunk.
        return $this->child_list[$child_id]['task_chunk'];
    }

    
    protected function
    fill_task_chunk_pool(){
        global $VMNC_MPL_task_prefetch_factor;
        global $VMNC_MPL_logger;

        # Calculate needed supply number.
        $max_n_tasks =
            ($this->n_children * $this->task_chunk_size)
            * $VMNC_MPL_task_prefetch_factor;

        # Ask the task supplier for new tasks.
        $new_tasks =
            $this->task_supplier_object->supply_new_tasks($max_n_tasks);

        $VMNC_MPL_logger->add(
            sprintf("Fetched %d new tasks", count($new_tasks)),
            VMNC_LOG_LEVEL_DEBUG
        );

        # This hack is necessary where an external database marks tasks
        # as done only after it is actively confirmed through execution
        # of the task. In such a system, a 'task in progress' can creep
        # up for second time in the 'new tasks'.
        if ($this->prevent_task_duplication){
            $unassigned_new_tasks = array_diff
                ($new_tasks, $this->tasks_in_progress);
                
            if (empty($unassigned_new_tasks))
                { throw new EmptyAfterDuplicateCheck(); }

            $n_dup = count($new_tasks) - count($unassigned_new_tasks);
            $VMNC_MPL_logger->add(
                "{$n_dup} duplicate tasks removed.",
                VMNC_LOG_LEVEL_DEBUG
            );

            $new_tasks = $unassigned_new_tasks;
        }

        # Divide the tasks into definite sized
        # chunks and then put them in the pool.
        $this->task_chunk_pool =
            array_merge($this->task_chunk_pool,
                array_chunk($new_tasks, $this->task_chunk_size));
    }


    protected function
    run_parent(){
        global $VMNC_MPL_logger, $VMNC_LOG_extra;

        $this->setup_clock();
        $request_queue = array();

        $should_die = False;
        $n_dead_children = 0;

        while (True){
            try {
                # Check if it's time for 'periodic polling'.
                $this->check_clock();

                if ($VMNC_LOG_extra > 0){
                    if ($n_dead_children){
                        $VMNC_MPL_logger->add(
                            "Number of dead children: {$n_dead_children}"
                            , VMNC_LOG_LEVEL_DEBUG
                        );
                    }
                }

                if ($should_die && ($n_dead_children >= $this->n_children)){
                    break; }

                if (count($request_queue) == 0){
                    $request_queue = $this->wait_for_new_request();
                    continue;
                }

                # Grab a request from 'request_queue'. The request is
                # not deleted until processing is confirmed, as lot's of
                # exceptions can happen in between.
                $request = $request_queue[0];

                $subject = $request['subject'];
                $child_id = $request['child_id'];

                switch ($subject){
                    case 'task_chunk': 
                        if ($should_die){
                            $this->tell_child_to_resign($child_id);
                            break;
                        }

                        try {
                            # Send a new 'task chunk' to the child.
                            $new_task_chunk =
                                $this->assign_new_task_chunk($child_id);

                            $this->send_task_chunk(
                                $child_id, $new_task_chunk);
                        }
                        catch (EmptyTaskChunkPool $excp){
                            # Refill the task chunk pool.
                            try { $this->fill_task_chunk_pool(); continue 2; }
                            catch (EmptyAfterDuplicateCheck $excp) {
                                # Since we could not properly handle current
                                # request, we ask the child to try again later.
                                $this->ask_to_try_later($child_id);
                            }
                        }

                        break;

                    case 'large_task_hint':
                        if ($should_die){
                            $this->tell_child_to_resign($child_id);
                            break;
                        }

                        $large_task_hint =
                            $this->task_supplier_object
                            ->supply_large_task_hint();

                        $this->send_large_task_hint(
                            $child_id, $large_task_hint);


                        break;

                    case 'task_status':
                        $this->task_supplier_object
                        ->process_task_status($request['task_status']);

                        break;

                    case 'resignment_letter':
                        if (isset($request['task_status'])){
                            $this->task_supplier_object
                            ->process_task_status($request['task_status']);
                        }

                        $this->send_resignment_letter_ack($child_id);

                        $n_dead_children += 1;
                        $should_die = True;

                        break;

                    case 'ask_to_resign':
                        $this->tell_child_to_resign($child_id);
                        break;

                    default:
                        throw new ThisShouldNeverHappenError();
                }

                array_shift($request_queue);
            }
            catch (WaitForNewRequestTimedOut $excp) { continue; }
            catch (TasksExhausted $excp){
                if (! $should_die){
                    $death_reason = "natural";
                    $should_die = True;
                }
            }
            catch (UnnaturalShutdown $excp) {
                if (! $should_die){
                    $death_reason = "unnatural";
                    $should_die = True;

                    foreach ($this->active_sockets as $child_id => $s){
                        $request_queue[] = array(
                            'subject' => 'ask_to_resign'
                            , 'child_id' => $child_id
                        );
                    }
                }
            }

        }

        return $death_reason;
    }
    
    protected function
    run(){
        # Prepare to create child processes.
        $this->task_supplier_object->do_before_start_of_split();
        
        # Create child processes.
        $this->create_children();
        
        # End of child process creation.
        $this->task_supplier_object->do_after_end_of_split();

        $fs = $this->run_parent();
        $fs = $this->task_supplier_object->modify_finish_status($fs);

        # Prepare to end child processes.
        $this->task_supplier_object->do_before_start_of_join();

        # All child processes are finished.
        $this->task_supplier_object->do_after_end_of_join();

        return $fs;
    }

    protected function
    cleanup(){
        # Forcefully resign the children.
        $this->resign_children(True);
    }

    public function
    start(){
        try {
            $this->setup_ipc();
            $fs = $this->run();
            $this->teardown_ipc();

            return $fs;
        }
        catch (\Exception $excp){
            $this->cleanup();
            throw $excp;
        }
    }
}


abstract class
ChildProcessManagerAbstract{
    public function
    __construct(
        $task_performer_class
    ){
        # Check if we are in a properly spawned child process.
        if (getenv('MPL_PROCESS_IS_CHILD') === False)
            { throw new NotAChildProcess(); }

        # Collect task_hint and process_hint.
        #-----------------------------------------------------------
        if (($th = getenv('MPL_TASK_HINT')) === False)
            { throw new MissingTaskHint(); }

        if (($ph = getenv('MPL_PROCESS_HINT')) === False)
            { throw new MissingProcessHint(); }

        if ((($this->task_hint = unserialize($th)) === False) &&
            ($th != serialize(False)))
            { throw new InvalidTaskHint(); }

        if ((($this->process_hint = unserialize($ph)) === False) &&
            ($ph != serialize(False)))
            { throw new InvalidProcessHint(); }
        #-----------------------------------------------------------

        $this->task_performer_class = $task_performer_class;
    }

    # This method should handle any setup required to
    # obtain an working ipc system. It may raise 
    # 'IpcError' exception in case of a failure.
    # The clues to setup IPC may be found in the 'process hint'.
    abstract protected function
    setup_ipc();

    # This method should cleanly shutdown the IPC mechanism.
    abstract protected function
    teardown_ipc();

    # cleanup() is called in case of an abnormal shutdown.
    abstract protected function
    cleanup();

    # This method should ask parent(through IPC) for new tasks,
    # and block until it receives a reply. If the the parent replies
    # with new tasks, the method should return them. If the parent
    # requests a 'resignment' instead, the method should raise a 
    # 'RequestedToResign' exception.
    abstract protected function
    get_new_task_chunk();

    abstract protected function
    get_large_task_hint();

    abstract protected function
    send_task_status($task_status);

    abstract protected function
    send_resignment_letter($task_status);

    abstract protected function
    wait_for_resignment_letter_ack();

    # This method is passed as a function argument to 'perform_task'
    # method of the 'task_perfomer_object'. It should check for
    # resignment requests sent (through IPC channel) by the parent
    # process and raise a 'RequestedToResign' on such occasion.
    abstract public function
    periodic_callback_B();

    protected function
    run(){
        $this->task_performer_object =
            new $this->task_performer_class(
                    $this->task_hint, $this->get_large_task_hint());

        for(;;) {
            try {
                $task_status =
                    $this->task_performer_object->perform_tasks(
                        $this->get_new_task_chunk(),

                        # This weird syntax is to pass a method as a callback.
                        array($this, 'periodic_callback_B')
                    );

                $this->send_task_status($task_status);
            }
            catch (RequestedToResign $excp){
                $this->send_resignment_letter($excp->task_status);
                $this->wait_for_resignment_letter_ack();
                break;
            }
        }
    }

    public function
    start(){
        try{
            $this->setup_ipc();
            $this->run();
            $this->teardown_ipc();
        }
        catch (\Exception $excp){
            $this->cleanup();
            throw $excp;
        }
    }
}
