<?php
namespace vimmaniac\mpl;

require_once(VMNC_MSH_PATH.'/lib/mpl/mpl.class.php');
require_once(VMNC_MSH_PATH.'/lib/soq/soq.class.php');

use vimmaniac\soq as soq;
use vimmaniac\logging as logging;

######### Finalized Exceptions (Should not be subclassed) ###############
final class SocketError extends IpcError {
    public function
    __construct($socket=Null, \Exception $previous=Null){
        $errcode = (is_null($socket))
            ? socket_last_error()
            : socket_last_error($socket);

        $errmsg = socket_strerror($errcode);
        parent::__construct(
            $errmsg, $errcode, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}

final class UnexpectedMsgReceived extends MplException {
    public function
    __construct($unexpected_msg, \Exception $previous=Null){
        $msg = "Received msg was unexpected - '''$unexpected_msg'''";
        parent::__construct($msg, Null, $previous, VMNC_LOG_LEVEL_WARN);
    }
}

final class ChildIdCrossCheckFailed extends MplException {
    public function
    __construct(\Exception $previous=Null){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}

final class FailedToCreateClientSocket extends MplException {
    public function
    __construct(\Exception $previous=Null){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}

final class InvalidChildIdReceived extends MplException {
    public function
    __construct(\Exception $previous=Null){
        parent::__construct(Null, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}
#########################################################################

$VMNC_MPL_socket_timeout = 1.0;
$VMNC_MPL_try_later_delay = 200000;
$VMNC_MPL_show_msg_trace = False;

class SoqDebug extends soq\Soq {
    public function send($msg, $block=False){
        global $VMNC_MPL_logger;

        $VMNC_MPL_logger->add(
            sprintf(
                "MSG_TRACE | %s >> %s"
                , getmypid(), $msg['subject']
            )
            , VMNC_LOG_LEVEL_DEBUG
        );

        $msg['pid'] = getmypid();
        return parent::send($msg, $block);
    }

    public function recv($block=False){
         global $VMNC_MPL_logger;

        $msg = parent::recv($block);
        $VMNC_MPL_logger->add(
            sprintf(
                "MSG_TRACE | %s << %s << %s"
                , getmypid(), $msg['subject'], $msg['pid']
            )
            , VMNC_LOG_LEVEL_DEBUG
        );

        return $msg;
    }
}

class ParentProcessManagerSoq
extends ParentProcessManagerAbstract {
    public function
    __construct(
        $host,
        $port,

        $task_supplier_object,
        $php_executable_path,
        $child_script_path,

        $n_children,
        $task_chunk_size,
        $periodic_polling_interval,

        $prevent_task_duplication = True
    ){
        global $VMNC_MPL_logger;

        $this->host = $host;
        $this->port = $port;
        $VMNC_MPL_logger->add(
            sprintf(
                "'%s' listening on - host: '''%s''', port:'''%s'''"
                , get_class($this), $this->host, $this->port
            )
            , VMNC_LOG_LEVEL_DEBUG
        );

        $this->n_soqs = 0;

        $this->map_socket_id_to_child_id = array();

        parent::__construct(
            $task_supplier_object,
            $php_executable_path,
            $child_script_path,

            $n_children,
            $task_chunk_size,
            $periodic_polling_interval,

            $prevent_task_duplication
        );
    }

    protected function
    send_task_chunk($child_id, $task_chunk){
        $this->child_list[$child_id]['soq']->send(
            array(
                'subject' => 'task_chunk'
                , 'task_chunk' => $task_chunk
            ),
            True
        );
    }

    protected function
    send_large_task_hint($child_id, $large_task_hint){
        $this->child_list[$child_id]['soq']->send(
            array(
                'subject' => 'large_task_hint'
                , 'large_task_hint' => $large_task_hint
            )
            , True
        );
    }

    protected function
    ask_to_try_later($child_id){
        global $VMNC_MPL_try_later_delay;

        $this->child_list[$child_id]['soq']->send(
            array(
                'subject' => 'try_later'
                , 'delay_usec' => $VMNC_MPL_try_later_delay
            )
            , True
        );
    }

    protected function
    tell_child_to_resign($child_id){
        $this->child_list[$child_id]['soq']->send(
            array('subject' => 'resign') , True
        );
    }

    protected function
    send_resignment_letter_ack($child_id){
        $this->child_list[$child_id]['soq']->send(
            array('subject' => "resignment_letter_ack")
            , True
        );
    }

    protected function
    setup_ipc(){
        # Create a server socket.
        $this->server_socket = socket_create(AF_INET, SOCK_STREAM, SOL_TCP);
        if (! is_resource($this->server_socket))
            { throw new SocketError(); }

        if (! socket_bind($this->server_socket, $this->host, $this->port))
            {throw new SocketError(); }

        # $this->port may be '0'(zero), which makes socket_bind to select
        # any freely available port and listen on it for incoming connections.
        # So, we want to know actual port the parent is listening, as we
        # need to tell our children to connect there.
        socket_getsockname($this->server_socket, $this->host, $this->port);

        if (! socket_listen($this->server_socket, $this->n_children))
            {throw new SocketError(); }

    }

    protected function
    create_a_soq(){
        global $VMNC_MPL_logger, $VMNC_MPL_show_msg_trace;

        # We should keep creating new socket connections and
        # 'Soq' objectes until all children are connected.
        if ($this->n_soqs >= $this->n_children) return;
        
        $csock = socket_accept($this->server_socket);
        if (! is_resource($csock))
            { throw new FailedToCreateClientSocket(); }

        # Accept a new connection.
        if ($VMNC_MPL_show_msg_trace)
            { $soq = new SoqDebug($csock); }
        else 
            { $soq = new soq\Soq($csock); }

        # Identify the child behind the connection by
        # reading the 'initiation msg'.
        #-----------------------------------------------------------
        $msg = $soq->recv(True);
        if ($msg['subject'] != 'init'){
            $VMNC_MPL_logger->add(
                'Client sent improper initiation message. Skipping.',
                VMNC_LOG_LEVEL_WARN
            );

            socket_close($csock);
            unset($soq);

            return;
        }

        $child_id = $msg['child_id'];
        if (! isset($this->child_list[$child_id]))
            { throw new InvalidChildIdReceived(); }

        $this->child_list[$child_id]['soq'] = $soq;
        $this->active_sockets[$child_id] = $soq->socket;
        $this->n_soqs += 1;
        #-----------------------------------------------------------
    }

    protected function
    wait_for_new_request(){
        global $VMNC_MPL_socket_timeout;

        if (! isset($this->soqs_initiated)){
            # Accept connections from all children, and 
            # create 'Soq' objects for all of them.
            while ($this->n_soqs < $this->n_children)
                { $this->create_a_soq(); }

            $this->empty_soqs = array();

            $this->soqs_initiated = True;
        }

        if (count($this->empty_soqs) >= $this->n_children){
            # Make all the sockets blocking.
            foreach ($this->active_sockets as $child_id => $s){
                $this->child_list[$child_id]['soq']->set_block(True);
            }

            $r = $this->active_sockets;
            $w = Null; $e = Null;

            $tv_sec = floor($VMNC_MPL_socket_timeout);
            $tv_usec = (int)(1000000 *($VMNC_MPL_socket_timeout - $tv_sec));

            $n = socket_select($r, $w, $e, $tv_sec, $tv_usec);

            if ($n === False) { throw new SocketError(); }
            else if ($n === 0) { throw new WaitForNewRequestTimedOut(); }
        }

        $requests = array();
        foreach ($this->child_list as $child_id => $child){
            if (! isset($this->active_sockets[$child_id])) continue;

            try {
                $msg = $child['soq']->recv(False);

                if (count($this->empty_soqs) > 0){
                    if (in_array($child_id, $this->empty_soqs)){
                        unset($this->empty_soqs[$child_id]);
                    }
                }
            }
            catch (soq\EmptyQueue $excp) {
                $this->empty_soqs[$child_id] = $child_id;
                continue; 
            }
            catch (soq\SocketConnectionClosed $excp){
                unset($this->active_sockets[$child_id]);
                continue;
            }

            # Cross check child id.
            if ($child_id != $msg['child_id'])
                { throw new ChildIdCrossCheckFailed(); }

            # Finally, list the child as requestee.
            $requests[] = $msg;
        }

        return $requests;
    }

    protected function
    supply_process_hint($child_id){
        return array(
            'host' => $this->host,
            'port' => $this->port,
            'child_id' => $child_id
        );
    }

    protected function
    teardown_ipc(){
        global $VMNC_MPL_logger;

        $VMNC_MPL_logger->add(
            "Waiting for child sockets to close ..."
            , VMNC_LOG_LEVEL_DEBUG
        );

        # Close client sockets.
        foreach ($this->active_sockets as $k => $socket){
            socket_set_block($socket); 
            while (True){
                $rc = socket_recv($socket, $dummy, 1024, 0);
                if (($rc === False) || ($dummy == '')) { break; }
            }
        }

        # Close server socket.
        @socket_close($this->server_socket);
    }
}

class ChildProcessManagerSoq
extends ChildProcessManagerAbstract {

    public function
    __construct(
        $task_executive_class
    ){
        parent::__construct($task_executive_class);

        $this->host = $this->process_hint['host'];
        $this->port = $this->process_hint['port'];
        $this->child_id = $this->process_hint['child_id'];
    }

    protected function
    setup_ipc(){
        global $VMNC_MPL_show_msg_trace;

        $csock = socket_create(AF_INET, SOCK_STREAM, SOL_TCP);
        if (! is_resource($csock)) { throw new SocketError(); }

        $rc = socket_connect($csock, $this->host, $this->port);
        if (! $rc) { throw new SocketError($csock); }

        if ($VMNC_MPL_show_msg_trace)
            { $this->soq = new SoqDebug($csock); }
        else
            { $this->soq = new soq\Soq($csock); }

        $this->soq->send(
            array(
                'subject' => 'init'
                , 'child_id' => $this->child_id
            )
            , True
        );
    }

    protected function
    teardown_ipc(){
        @socket_close($this->soq->socket);
    }

    protected function
    cleanup(){
        $this->teardown_ipc();
    }

    protected function
    get_new_task_chunk(){
        while (True){
            $this->soq->send(
                array(
                    'subject' => 'task_chunk'
                    , 'child_id' => $this->child_id
                )
                , True
            );

            $msg = $this->soq->recv(True);
            switch ($msg['subject']){
                case 'try_later':
                    usleep($msg['delay_usec']);
                    continue;
                    
                case 'task_chunk': 
                    return $msg['task_chunk'];

                case 'resign':
                    throw new RequestedToResign();

                default:
                    throw new UnexpectedMsgReceived(print_r($msg, True));
            }
        }
    }

    protected function
    get_large_task_hint(){
        $this->soq->send(
            array(
                'subject' => 'large_task_hint'
                , 'child_id' => $this->child_id
            )
            , True
        );

        $msg = $this->soq->recv(True);
        switch ($msg['subject']){
            case 'large_task_hint': 
                return $msg['large_task_hint'];

            case 'resign':
                throw new RequestedToResign();

            default:
                throw new UnexpectedMsgReceived(print_r($msg, True));
        }
    }

    protected function
    send_task_status($task_status){
        $this->soq->send(
            array(
                'subject' => 'task_status',
                'child_id' => $this->child_id,
                'task_status' => $task_status
            ),
            True
        );
    }

    protected function
    send_resignment_letter($task_status){
        $this->soq->send(
            array(
                'subject' => 'resignment_letter'
                , 'child_id' => $this->child_id
                , 'task_status' => $task_status
            )
            , True
        );

    }

    protected function
    wait_for_resignment_letter_ack(){
        do { $msg = $this->soq->recv(True); }
        while ($msg['subject'] != 'resignment_letter_ack');
    }

    public function
    periodic_callback_B(){
        try {
            $msg = $this->soq->recv(False);

            if ($msg['subject'] == 'resign')
                { throw new RequestedToResign(); }
            else
                { throw new UnexpectedMsgReceived(print_r($msg, True)); }
        }
        catch (soq\EmptyQueue $excp) {}
    }
}
