<?php
/**
 * Soq : A duplex pipe-like thinlayer on top of sockets
 *           with message boundary detection.
 *
 * copyright (C) 2011 Titon Barua <titanix88@gmail.com>
 * license: This software is licensed under
 *          GNU Lesser General Public License.
 *          See LICENSE.txt for the full license description.
 *
 * version 1.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 **/

namespace vimmaniac\soq;

require_once(VMNC_MSH_PATH.'/lib/util/logging.php');
require_once(VMNC_MSH_PATH.'/lib/util/base_exception.php');

use vimmaniac\logging as logging;
if (! isset($VMNC_SOQ_logger))
    { $VMNC_SOQ_logger = new logging\CreateLogger('SOQ'); }

if (! defined('SOCKET_EAGAIN'))
    { define('SOCKET_EAGAIN', SOCKET_EWOULDBLOCK); }
else if (! defined('SOCKET_EWOULDBLOCK'))
    { define('SOCKET_EWOULDBLOCK', SOCKET_EAGAIN); }

$VMNC_SOQ_default_start_marker = '~S:';
$VMNC_SOQ_default_end_marker = ':E~';
$VMNC_SOQ_default_avg_msg_size = 2048;
$VMNC_SOQ_default_incoming_chunk_size = 2048;
$VMNC_SOQ_default_outgoing_chunk_size = 2048;
$VMNC_SOQ_default_send_in_chunks = False;

$VMNC_SOQ_fix_windows_quirk_1 = True;
$VMNC_SOQ_fix_windows_quirk_2 = True;

if (! defined('VMNC_OS_IS_WINDOWS')){
    define(
        'VMNC_OS_IS_WINDOWS'
        , (strtolower(substr(PHP_OS, 0, 3)) == 'win')
    );
}

class SoqError extends \vimmaniac\BaseException {
    public function
    __construct(
        $message=Null,
        $code=Null,
        \Exception $previous=Null,
        $log_level=VMNC_LOG_LEVEL_DEBUG
    ){
        global $VMNC_SOQ_logger;

        parent::__construct($message, $code, $previous);
        $VMNC_SOQ_logger->add($this->__toString(), $log_level);
    }
}

class PackingError extends SoqError {
    public function
    __construct(
        $message='Failed to pack outgoing message.',
        $code=Null,
        \Exception $previous=Null,
        $log_level=VMNC_LOG_LEVEL_ERROR
    ){
        parent::__construct($message, $code, $previous, $log_level);
    }
}


class SerializationError extends PackingError {
    public function
    __construct(
        $message='Failed to serialize outgoing message.',
        $code=Null,
        \Exception $previous=Null,
        $log_level=VMNC_LOG_LEVEL_ERROR
    ){
        parent::__construct($message, $code, $previous, $log_level);
    }
}


class UnpackingError extends SoqError {
    public function
    __construct(
        $message='Failed to unpack incoming message.',
        $code=Null,
        \Exception $previous=Null,
        $log_level=VMNC_LOG_LEVEL_ERROR
    ){
        parent::__construct($message, $code, $previous, $log_level);
    }

}


class UnserializationError extends UnpackingError {
    public function
    __construct(
        $message='Failed to unserialize incoming message.',
        $code=Null,
        \Exception $previous=Null,
        $log_level=VMNC_LOG_LEVEL_ERROR
    ){
        parent::__construct($message, $code, $previous, $log_level);
    }
}


class AuthenticationError extends UnpackingError {
    public function
    __construct(
        $message='The received message is not properly authenticated.',
        $code=Null,
        \Exception $previous=Null,
        $log_level=VMNC_LOG_LEVEL_ERROR
    ){
        parent::__construct($message, $code, $previous, $log_level);
    }
}

########## Finalized Exceptions (Should not be subclassed) ###############
final class SocketError extends SoqError {
    public function
    __construct($socket, \Exception $previous=Null){
        $code = socket_last_error($socket);
        $message = socket_strerror($code);

        parent::__construct(
            $message, $code, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}

final class EmptyQueue extends  SoqError {
    public function
    __construct(\Exception $previous=Null){
        $msg = 'Incoming queue is empty.';
        parent::__construct($msg, Null, $previous, VMNC_LOG_LEVEL_DEBUG);
    }
}

final class UnfinishedSend extends SoqError {
    public function
    __construct(\Exception $previous=Null){
        $msg = 'Last send operation did not finish.';
        parent::__construct($msg, Null, $previous, VMNC_LOG_LEVEL_WARN);
    }
}

final class SocketBlockModeChangeFailed extends SoqError {
    public function
    __construct(\Exception $previous=Null){
        $msg = 'Failed to change block mode.';
        parent::__construct($msg, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}

final class SocketConnectionClosed extends SoqError {
    public function
    __construct($socket, \Exception $previous=Null){
        $msg = 'Peer has closed connection.';
        parent::__construct($msg, Null, $previous, VMNC_LOG_LEVEL_ERROR);
    }
}
##########################################################################

class Soq {
    public $socket = Null;

    public $average_message_size;
    public $incoming_chunk_size;
    public $outgoing_chunk_size;
    public $send_in_chunks;
 
    protected $incoming_queue; 
    protected $incoming_buffer;
    protected $outgoing_buffer;

    protected $start_marker;
    protected $start_marker_length;
    
    protected $end_marker;
    protected $end_marker_length;

    protected $socket_is_blocking;
   
    public function
    __construct(
        $socket,

        $start_marker = Null,
        $end_marker = Null,

        $average_message_size = Null,
        $incoming_chunk_size = Null,
        $outgoing_chunk_size = Null,
        $send_in_chunks = Null
    ){
        global
            $VMNC_SOQ_default_start_marker,
            $VMNC_SOQ_default_end_marker,
            $VMNC_SOQ_default_avg_msg_size,
            $VMNC_SOQ_default_incoming_chunk_size,
            $VMNC_SOQ_default_outgoing_chunk_size,
            $VMNC_SOQ_default_send_in_chunks,
            $VMNC_SOQ_logger;

        $this->socket = $socket;

        $this->start_marker =
            (is_null($start_marker))
            ? $VMNC_SOQ_default_start_marker : $start_marker;
        $this->start_marker_length = strlen($this->start_marker);

        $this->end_marker =
            (is_null($start_marker))
            ? $VMNC_SOQ_default_end_marker : $end_marker;
        $this->end_marker_length = strlen($this->end_marker);

        $this->average_message_size =
            (is_null($average_message_size))
            ? $VMNC_SOQ_default_avg_msg_size : $average_message_size;

        $this->incoming_chunk_size =
            (is_null($incoming_chunk_size))
            ? $VMNC_SOQ_default_incoming_chunk_size : $incoming_chunk_size;

        $this->outgoing_chunk_size =
            (is_null($outgoing_chunk_size))
            ? $VMNC_SOQ_default_outgoing_chunk_size : $outgoing_chunk_size;
            
        $this->send_in_chunks =
            (is_null($send_in_chunks))
            ? $VMNC_SOQ_default_send_in_chunks : $send_in_chunks;

        $VMNC_SOQ_logger->add(
            sprintf(
                "Instance of class '%s' created with parameters -"
                ." start marker: '''%s'''"
                .", end_marker: '''%s'''"
                .", average_message_size: '''%s'''"
                .", incoming_chunk_size: '''%s'''"
                .", outgoing_chunk_size: '''%s'''"
                .", send_in_chunks: '''%s'''"
                , get_class($this)
                , $this->start_marker
                , $this->end_marker
                , $this->average_message_size
                , $this->incoming_chunk_size
                , $this->outgoing_chunk_size
                , var_export($this->send_in_chunks, True)
            )
            , VMNC_LOG_LEVEL_DEBUG
        );

        # Setup a fifo queue for incoming messages.
        $this->incoming_queue = new \SplQueue();
        $this->incoming_queue->setIteratorMode(\SplQueue::IT_MODE_DELETE);
 
        $this->reset_mss_vars();
    }


    public function
    set_block($make_blocking){
        global $VMNC_SOQ_logger, $VMNC_LOG_extra;
        
        if ($VMNC_LOG_extra > 10){
            $c = ($this->socket_is_blocking) ? 'blocking' : 'non-blocking';
            $f = ($make_blocking) ? 'blocking' : 'non-blocking';
            $VMNC_SOQ_logger->add(
                "Changing socket mode: $c => $f", VMNC_LOG_LEVEL_DEBUG);
        }

        if ($make_blocking){
            if (! socket_set_block($this->socket))
                { throw new SocketBlockModeChangeFailed(); }
            $this->socket_is_blocking = True;
        }
        else {
            if (! socket_set_nonblock($this->socket))
                { throw new SocketBlockModeChangeFailed(); }
            $this->socket_is_blocking = False;
        }
    }


    protected function
    pack_message($msg){
        global $VMNC_SOQ_logger, $VMNC_LOG_extra;

        #TODO: check whether serialization failed or not.
        $smsg = serialize($msg);

        if ($VMNC_LOG_extra > 5){
            $VMNC_SOQ_logger->add(
                sprintf(
                    "Raw outgoing msg: <%s>"
                    .", Packed outgoing msg: <%s>"
                    , var_export($msg, True) , $smsg
                )
                , VMNC_LOG_LEVEL_DEBUG
            );
        }

        return $smsg;
    }


    protected function
    unpack_message($serialized_msg){
        global $VMNC_SOQ_logger, $VMNC_LOG_extra;

        $msg = unserialize($serialized_msg);
        if (($msg === False) && (serialize(False) != $unserialized_msg)){
            throw new UnserializationError();
        }

        if ($VMNC_LOG_extra > 5){
            $VMNC_SOQ_logger->add(
                sprintf(
                    "Packed incoming msg: <%s>"
                    .", Unpacked incoming msg: <%s>"
                    , $serialized_msg, var_export($msg, True)
                )
                , VMNC_LOG_LEVEL_DEBUG
            );
        }

        return $msg;
    }


    protected function
    send_($block){
        global $VMNC_SOQ_fix_windows_quirk_2;

        # ==========> Windows Specific Hack <==============
        # send_ just seem to block without any reason
        # in windows, whenever started through 'mpl'.
        # Why the hell does it happen? I don't know.
        # The solution is as strange as the problem.
        # If *anything* is 'echo'ed to the standard output,
        # it behaves normally. 
        if (VMNC_OS_IS_WINDOWS && $VMNC_SOQ_fix_windows_quirk_2)
            { echo "\0"; }

        if ($block){
            if (! $this->socket_is_blocking)
                { $this->set_block(True); }
        }
        else {
            if ($this->socket_is_blocking)
                { $this->set_block(False); }
        }

        while (True){
            # Grab a packet(chunk of data) from outgoing buffer.
            $pckt =
                ($this->send_in_chunks)
                ? substr($this->outgoing_buffer, 0, $this->outgoing_chunk_size)
                : $this->outgoing_buffer;

            # Finish if outgoing buffer is empty.
            if (strlen($pckt) === 0) { break; }

            # Send the packet to the peer.
            $n_sent_bytes = socket_write($this->socket, $pckt);

            # Handle errors.
            if ($n_sent_bytes === False){
                switch (socket_last_error($this->socket)){

                    case SOCKET_EWOULDBLOCK:
                    case SOCKET_EAGAIN:

                        # For blocking mode, try again.
                        if ($this->socket_is_blocking) { continue; }

                        # For non-blocking mode, raise an exception to
                        # notify the caller.
                        else { throw new UnfinishedSend(); }

                    default:
                        throw new SocketError($this->socket);
                }
            }

            # socket_write failed to transmit any data at all.
            # For non-blocking mode, notify the caller.
            if (($n_sent_bytes === 0) && (! $this->socket_is_blocking)){
                throw new UnfinishedSend();
            }

            # Discard the the portion of the outgoing buffer
            # that has just been sent to the peer.
            $this->outgoing_buffer = substr(
                $this->outgoing_buffer, $n_sent_bytes);
        }
    }

    public function
    send($msg, $block=False){
        global $VMNC_SOQ_logger, $VMNC_LOG_extra;

        $smsg = $this->pack_message($msg);

        # Please consult separate_msg_from_incoming_buffer()
        # for protocol description.
        $this->outgoing_buffer .= (
            $this->start_marker
            ."[".(string)strlen($smsg)."]"
            .$smsg
            .$this->end_marker
        );

        if ($VMNC_LOG_extra > 15){
            $VMNC_SOQ_logger->add(
                "Outgoing buffer: <".$this->outgoing_buffer.">"
                , VMNC_LOG_LEVEL_DEBUG
            ); 
        }

        $this->send_($block);
    }

    protected function
    recv_($block){
        if ($block){
            if (! $this->socket_is_blocking)
                { $this->set_block(True); }
        }
        else {
            if ($this->socket_is_blocking)
                { $this->set_block(False); }
        }

        $incoming_data = "";
        $incoming_data_size = 0;

        while (True){
            global $VMNC_SOQ_fix_windows_quirk_1;

            $n_bytes = @socket_recv(
                $this->socket, $received_msg, $this->incoming_chunk_size, 0);

            if ($n_bytes === False){
                $sle = socket_last_error($this->socket);

                # SOCKET_EWOULDBLOCK and SOCKET_EAGAIN can only araise when
                # socket is set as non-blocking and there is not enough data
                # to read. Nothing to do but to break out.
                if (($sle == SOCKET_EWOULDBLOCK) ||
                    ($sle == SOCKET_EAGAIN))
                    { break; }
                else {
                    # ==> WINDOWS SPECIFIC HACK <==
                    # In case of a non-blocking socket, if the socket does not
                    # have any data to be read, a False is returned, with a 
                    # socket error of zero [0] "The operation completed
                    # successfully." Very very odd, if you ask me! %(
                    if  ( ($sle === 0)
                          && VMNC_OS_IS_WINDOWS
                          && $VMNC_SOQ_fix_windows_quirk_1 ) { break; }

                    throw new SocketError($this->socket);
                }
            }

            # socket_recv() returns 0 only
            # if the peer has closed connection.
            else if ($n_bytes === 0)
                { throw new SocketConnectionClosed($this->socket); }

            $incoming_data .= $received_msg;
            $incoming_data_size += strlen($received_msg);

            # After receiving the first packet, disable blocking.
            if ($this->socket_is_blocking) { $this->set_block(False); }

            # If incoming data is large enough, we should break out
            # and try to decode a message from it. If done otherwise,
            # situation can arise where the sending side thrashes the
            # receiving side with large amounts of messages and the
            # receiving side keeps gathering them(without decoding) forever,
            # eventually running out of memory.
            if ($incoming_data_size >=  $this->average_message_size)
                { break; }
        }

        $this->incoming_buffer .= $incoming_data;

        return ($incoming_data_size === 0) ? False : True;

    }


    protected function
    prepare_for_new_msg($discard_size=0){
            if ($discard_size === 0){

                # A zero discard size means "discard message in progress".
                # Since every message starts with the start_marker, removing
                # upto and including the start_marker of the message 
                # discards and prevents it from being matched again.

                if (($this->mss_start_marker_found) &&
                    (! is_null($this->mss_start_marker_pos))){
                    $this->incoming_buffer = substr(
                        $this->incoming_buffer,
                        $this->mss_start_marker_pos
                            + $this->start_marker_length);
                }
            }
            else {
                $this->incoming_buffer = substr(
                    $this->incoming_buffer,
                    $discard_size);
            }

            $this->reset_mss_vars();
    }


    protected function
    reset_mss_vars(){
        # mss => Message Separation States.

        $this->mss_start_marker_found = False;
        $this->mss_start_marker_pos = Null;

        $this->mss_msg_length_found = False;
        $this->mss_msg_length = Null;

        $this->mss_msg_pos = Null;

    }


    protected function
    separate_msg_from_incoming_buffer(){

        # In this function, message separation states are reset (
        # by calling reset_msg_separation_state() function) only
        # when it is evident that part of the message is invalid.
        #
        # The the following comments -
        #   Start Marker: "[START]"
        #   End Marker:   "[END]"
        #
        # An example of a received message -
        #   [START][18]<serialized hello>[END]
        #   |     ||  ||                ||   |
        #   +--+--+|  |+--------+-------+|   |
        #      |   |  |         |        |   |
        #      |   +--+         |        +---+
        #      |      |         |        |
        #      |      |         |        |
        #      |      |         |        |
        #      |      |         |    End Marker
        #      |      |         |
        #      |      |   Serialized Message
        #      |      |
        #      |  Length of the serialized message.
        #      |  The square brackets are part of the protocol.
        #      |
        #   Start Marker
        #

        $iblen = strlen($this->incoming_buffer);

        if (! $this->mss_start_marker_found){

            ######################################
            # Invalid: "[ST", "[START"
            # Valid:   "[START]---", "----------"
            #========================================================\
            if ($iblen < $this->start_marker_length) { return False; }
            #========================================================/

            #############################
            # Invalid: "----------[STA"
            # Valid:   "------[START]--"
            #=======================================================\
            $x = strpos($this->incoming_buffer, $this->start_marker);
            if ($x === False) { return False; }
            #=======================================================/

            $this->mss_start_marker_found = True;
            $this->mss_start_marker_pos = $x;
        }


        if (! $this->mss_msg_length_found){
            $x = ($this->mss_start_marker_pos
                  + $this->start_marker_length);

            #############################################
            # Invalid: "-----[START][", "---[START]--"
            # Valid:   "---[START][4]", "---[START]-----"
            #======================================\
            if ($iblen < ($x + 3)) { return False; }
            #======================================/

            ##########################
            # Invalid: "[START]XY--"
            # Valid:   "[START][---"
            #=======================================\
            if ($this->incoming_buffer[$x] != "["){
                $this->prepare_for_new_msg();
                return False;
            }
            #=======================================/

            $y = strpos($this->incoming_buffer, "]", $x + 1);
            if ($y === False){
                #####################################
                # Invalid: "------[START][12432XVB"
                # Valid:   "------[START][12343444"
                #========================================================\
                if (! ctype_digit(substr($this->incoming_buffer, $x + 1)))
                    { $this->prepare_for_new_msg(); }
                return False;
                #========================================================/
            }

            $msg_length = substr($this->incoming_buffer, $x + 1, $y - $x - 1);
            ####################################
            # Invalid: "-----[START][12NG]---"
            # Valid:   "-----[START][15]-----"
            #========================================\
            if (! ctype_digit($msg_length)){
                $this->prepare_for_new_msg();
                return False;
            }
            #========================================/

            $this->mss_msg_length_found = True;
            $this->mss_msg_length = (int)$msg_length;
            $this->mss_msg_pos = $y + 1;
        }

        $msg_with_end_marker = substr(
            $this->incoming_buffer, $this->mss_msg_pos);


        # Check if we have received the minimumpackingata.
        if ( strlen($msg_with_end_marker)
             < ($this->mss_msg_length + $this->end_marker_length))
            { return False; }


        $end_marker_etc = substr(
            $msg_with_end_marker, $this->mss_msg_length);


        # Check if end marker is in it's expected position.
        if (! (strpos($end_marker_etc, $this->end_marker) === 0)){
            $this->prepare_for_new_msg();
            return False;
        }


        # Separate the serialized message.
        $smsg = substr($msg_with_end_marker, 0, $this->mss_msg_length);


        # Get ready for matching a new message.
        $this->prepare_for_new_msg(
            $this->mss_msg_pos
            + $this->mss_msg_length
            + $this->end_marker_length);

        # Push the new msg into the queue.
        try {
            $this->incoming_queue->enqueue($this->unpack_message($smsg));
            return True;
        }
        catch (UnpackingError $e) { return False; }
    }


    protected function
    separate_all_msg_from_incoming_buffer(){
        $msg_separated = False;

        while (True){
            if (! $this->separate_msg_from_incoming_buffer()) break;
            $msg_separated = True;
        }

        # $msg_separated reports whether at least
        # one message has been separated or not.
        return $msg_separated;
    }


    public function
    update_incoming(){
        # This function is useful to check for incoming messages,
        # without poping one from the incoming queue. This might
        # be useful in high bandwidth applications where system
        # tcp buffer can get filled faster than the application
        # is capable of processing the messages.
        if ($this->recv_(False) === False) { return; }
        return $this->separate_all_msg_from_incoming_buffer();
    }


    public function
    flush_outgoing($block=True){
        # Complementary to update_incoming(), this function can be
        # used to try to send the messages already present in the
        # outgoing_buffer. This can be specially handy when a previous
        # send becomes unsuccessful with a UnfinishedSend exception.
        $this->send_($block);
    }


    public function
    recv($block=False){
        do {
            if ($this->incoming_queue->count() > 0)
                { return $this->incoming_queue->dequeue(); }

            # Try to grab a new message from the tcp buffer,
            # only if incoming_queue is completely empty.
            else {
                if ($this->recv_($block) == False) { continue; }
                if ($this->separate_all_msg_from_incoming_buffer() == False)
                    { continue; }

                return $this->incoming_queue->dequeue();
            }
        }
        while ($block);

        # For non-blocking mode, notify the caller about emtpy queue.
        throw new EmptyQueue();
    }
}
