<?php
namespace vimmaniac\msh;

define('VMNC_MSH_PATH', dirname(__DIR__));
define('VMNC_IEM_ROOT', dirname(dirname(dirname(dirname(__DIR__)))));
define('VMNC_MSH_CONFIG_FILE', VMNC_MSH_PATH.'/msh_config.ini');

require_once(VMNC_MSH_PATH.'/lib/util/logging.php');
require_once(VMNC_MSH_PATH.'/lib/util/base_exception.php');
require_once(VMNC_MSH_PATH.'/lib/mpl/mpl.class.php');
require_once(VMNC_MSH_PATH.'/lib/mpl/mpl_soq.class.php');
require_once(VMNC_MSH_PATH.'/lib/soq/soq.class.php');

require_once(VMNC_MSH_PATH.'/src/msh.class.php');
require_once(VMNC_MSH_PATH.'/src/jobs_send_msh.php');

$VMNC_MSH_child_script_path = VMNC_MSH_PATH.'/src/msh_child_script.php';

if (! isset($VMNC_MSH_logger))
    { $VMNC_MSH_logger = new \vimmaniac\logging\CreateLogger('MSH'); }

function read_config_file($config_file){
    global $VMNC_MSH_logger;

    $VMNC_MSH_logger->add(
        "Reading configurations from file: \"{$config_file}\".",
        VMNC_LOG_LEVEL_DEBUG
    );

    # Check if the file exists and is readable.
    if (! (file_exists($config_file)
           && is_readable($config_file)))
    {
        $VMNC_MSH_logger->add(
            "Failed to read configuration file.",
            VMNC_LOG_LEVEL_ERROR
        );

        exit(1);
    }

    # Parse configuration file.
    $configs = parse_ini_file($config_file, False, INI_SCANNER_RAW);
    if ($configs === False){
        $VMNC_MSH_logger->add(
            "Syntax error in configuration file.",
            VMNC_LOG_LEVEL_ERROR
        );

        exit(1);
    }

    return  $configs;
}

function _parse_bool($txt, $default){
    switch (strtolower(trim($txt))){
        case 'yes':
        case 'on':
        case '1':
        case 'true':
            return True;

        case 'no':
        case 'off':
        case '0':
        case 'false':
            return False;

        default:
            return $default;
    }
}

function _parse_int($txt, $min=Null, $max=Null){
    $v = intval($txt);
    if (! is_null($min)) { $v = ($v < $min) ? $min : $v; }
    if (! is_null($max)) { $v = ($v > $max) ? $max : $v; }

    return $v;
}

function _parse_float($txt, $min=Null, $max=Null){
    $v = floatval($txt);
    if (! is_null($min)) { $v = ($v < $min) ? $min : $v; }
    if (! is_null($max)) { $v = ($v > $max) ? $max : $v; }

    return $v;
}

if (! isset($VMNC_MSH_config_file_parsed)){
    $cvs = read_config_file(VMNC_MSH_CONFIG_FILE);

    $default_cvs = array(
        'enable_msh_addon' => 'Yes'
        , 'number_of_worker_process' => '4'
        , 'php_cli_path' => '"php"'
        , 'fix_iem_resend_bug' => 'Yes'
        , 'fix_php_socket_bug_1' => 'Yes'
        , 'fix_php_socket_bug_2' => 'Yes'
        , 'fix_php_proc_open_bug_1' => 'Yes'
        , 'fix_php_proc_open_bug_2' => 'Yes'
        , 'work_chunk_size' => '30'
        , 'work_prefetch_factor' => '1.5'
        , 'early_database_cleanup' => 'On'
        , 'early_database_cleanup_granularity' => '10'
        , 'parent_poll_period' => '1.0'
        , 'child_poll_period' => '1.0'
        , 'ipc_host' => '127.0.0.1'
        , 'ipc_port' => '0'
        , 'log_level' => 'info'
        , 'log_extra' => '0'
    );

    $cvs = array_merge($default_cvs, $cvs);

    $VMNC_MSH_logger->add(
        sprintf(
            "Parsed configuration parameters -\nenable_msh_addon = %s\nnumber_of_worker_process = %s\nphp_cli_path = %s\nfix_iem_resend_bug = %s\nfix_php_socket_bug_1 = %s\nfix_php_socket_bug_2 = %s\nfix_php_proc_open_bug_1 = %s\nfix_php_proc_open_bug_2 = %s\nwork_chunk_size = %s\nwork_prefetch_factor = %s\nearly_database_cleanup = %s\nearly_database_cleanup_granularity = %s\nparent_poll_period = %s\nchild_poll_period = %s\nipc_host = %s\nipc_port = %s\nlog_level = %s\nlog_extra = %s"
            , $cvs['enable_msh_addon']
            , $cvs['number_of_worker_process']
            , $cvs['php_cli_path']
            , $cvs['fix_iem_resend_bug']
            , $cvs['fix_php_socket_bug_1']
            , $cvs['fix_php_socket_bug_2']
            , $cvs['fix_php_proc_open_bug_1']
            , $cvs['fix_php_proc_open_bug_2']
            , $cvs['work_chunk_size']
            , $cvs['work_prefetch_factor']
            , $cvs['early_database_cleanup']
            , $cvs['early_database_cleanup_granularity']
            , $cvs['parent_poll_period']
            , $cvs['child_poll_period']
            , $cvs['ipc_host']
            , $cvs['ipc_port']
            , $cvs['log_level']
            , $cvs['log_extra']
        )
        ,
        ((getenv('MPL_PROCESS_IS_CHILD'))
            ? VMNC_LOG_LEVEL_DEBUG
            : VMNC_LOG_LEVEL_INFO
        )
    );

    $VMNC_MSH_enabled = _parse_bool($cvs['enable_msh_addon'], True);

    $VMNC_MSH_n_children =
        _parse_int($cvs['number_of_worker_process'], 1);

    $VMNC_MSH_php_cli_binary_path = $cvs['php_cli_path'];

    $VMNC_MSH_fix_iem_resend_bug =
        _parse_bool($cvs['fix_iem_resend_bug'], True);


    $VMNC_SOQ_fix_windows_quirk_1 =
        _parse_bool($cvs['fix_php_socket_bug_1'], True);

    $VMNC_SOQ_fix_windows_quirk_2 =
        _parse_bool($cvs['fix_php_socket_bug_2'], True);

    $VMNC_MPL_fix_windows_quirk_1 =
        _parse_bool($cvs['fix_php_proc_open_bug_1'], True);

    $VMNC_MPL_fix_windows_quirk_2 =
        _parse_bool($cvs['fix_php_proc_open_bug_2'], True);


    $VMNC_MSH_task_chunk_size = 
        _parse_int($cvs['work_chunk_size'], 1);


    $VMNC_MSH_task_prefetch_factor =
        _parse_float($cvs['work_prefetch_factor'], 0.5);

    $VMNC_MSH_enable_periodic_dbcleanup =
        _parse_bool($cvs['early_database_cleanup'], True);

    $VMNC_MSH_dbcleanup_granularity = 
        _parse_int($cvs['early_database_cleanup_granularity'], 1);

    $VMNC_MSH_parent_poll_period =
        _parse_float($cvs['parent_poll_period'], 0.0);

    $VMNC_MSH_child_poll_period =
        _parse_float($cvs['child_poll_period'], 0.0);

    $VMNC_MSH_ipc_host = $cvs['ipc_host'];

    $VMNC_MSH_ipc_port =
        _parse_int($cvs['ipc_port'], 0, 65535);

    $log_level = VMNC_LOG_LEVEL_INFO;
    if (isset($cvs['log_level'])){
        switch (strtolower($cvs['log_level'])){
             case 'error':
                $log_level = VMNC_LOG_LEVEL_ERROR;
                break;
            case 'warn':
                $log_level = VMNC_LOG_LEVEL_WARN;
                break;
            case 'debug':
                $log_level = VMNC_LOG_LEVEL_DEBUG;
                break;
            case 'info':
            default:
                $log_level = VMNC_LOG_LEVEL_INFO;
                break;
        }
    }

    $VMNC_MSH_logger->log_level =
        $VMNC_MPL_logger->log_level =
            $VMNC_SOQ_logger->log_level = $log_level;


    $VMNC_LOG_extra =
        (isset($cvs['log_extra']))
        ? _parse_int($cvs['log_extra'], 0, 100) : 0;

    $VMNC_MSH_config_file_parsed = True;
}

